/*
 * Copyright (c) 2002 Tomas Svensson <ts@unix1.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> 
                
#include <netdb.h>
        
#include "lamefwd.h"
#include "config.h"
#include "io.h"
#include "network2.h"

extern int debug;

#define LISTEN_BACKLOG 25

void close_ud(struct ud *ud)
{
	close(ud->user_fd);
	if (ud->serv_fd != -1)
		close(ud->serv_fd);
	ud->connected = CONN_NO;
	ud->user_fd = ud->serv_fd = -1;
}

int find_free_slot(struct ud *ud, int max_users) {

        /* Returns the index of the first free ud slot, otherwise
           it returns -1 */

        int i;
  
        for (i = 0; i < max_users; i++)
                if (ud++->user_fd == -1) return i;
  
        return -1;
}

void io_setup(struct fwdent *fwdent, int ents, struct ud *ud, int cfg_max_users)
{
	int i, result;
	for (i = 0; i < ents; i++) {
		fwdent[i].listenfd = setup_listen(LISTEN_BACKLOG,
		    fwdent[i].localport, (fwdent[i].proto ==
		    FWD_TCP) ? SOCK_STREAM : SOCK_DGRAM);
		if (fwdent[i].proto == FWD_UDP)
			fwdent[i].dgramfd = setup_connect(SOCK_DGRAM,
                            (struct sockaddr*)&fwdent[i].sin, sizeof(fwdent[i].sin), &result);
	}
	for (i=0; i < cfg_max_users; i++)
		ud[i].user_fd = -1;
}
void io_main(struct fwdent *fwdent, int ents, struct ud *ud, int cfg_max_users)
{
	int	i, n, newsock, nfds;
	fd_set	cset, rset, wset;
	struct sockaddr sa;
	socklen_t slen;
	int	bytes, bytesW, idx, result;
	char dgrambuf[2048], fakebuf[1];
	
	bytesW = 0;
  nfds = getdtablesize();
	if (FD_SETSIZE < nfds)
		nfds = FD_SETSIZE - 1;
	FD_ZERO(&cset);
	for (i = 0; i < ents; i++) {
		FD_SET(fwdent[i].listenfd, &cset);
		if (fwdent[i].proto == FWD_UDP)
			FD_SET(fwdent[i].dgramfd, &cset);
	}
	
	for (;;) {
		rset = cset;
		FD_ZERO(&wset);

		for(i = 0; i < cfg_max_users ; i++) {
                        if (ud[i].user_fd != -1) {
                                if (ud[i].connected == CONN_IN_PROG) {
                                        FD_SET(ud[i].serv_fd, &rset);
                                        FD_SET(ud[i].serv_fd, &wset);
                                } else if (ud[i].connected == CONN_YES) {
					if (ud[i].u2s_i != ud[i].u2s_o)
                                		FD_SET(ud[i].serv_fd, &wset);
                        		if (ud[i].s2u_i != ud[i].s2u_o)
                                		FD_SET(ud[i].user_fd, &wset);
					if (ud[i].s2u_i < &ud[i].s2u_buf[S2U_SIZE])
                                        	FD_SET(ud[i].serv_fd, &rset);
					if (ud[i].u2s_i < &ud[i].u2s_buf[U2S_SIZE])
						FD_SET(ud[i].user_fd, &rset);
				}
			}
		}

		n = select(nfds + 1, &rset, &wset, NULL, NULL);	

		if (n < 0) {
			if (errno == EINTR)
				continue;
			else
				sys_err("select error");
		}

		/* check if anyone has connected */

		for (i = 0; i < ents; i++) {
			
			if (fwdent[i].proto == FWD_UDP)
				if (FD_ISSET(fwdent[i].dgramfd, &rset)) {
					do {
						bytes = read(fwdent[i].dgramfd, dgrambuf, sizeof(dgrambuf));
						if (bytes >= 0)
							bytesW = sendto(fwdent[i].listenfd, dgrambuf, bytes,
							    0, &fwdent[i].sender, fwdent[i].senderlen);
					} while (bytes >= 0 && bytesW == bytes);
				}
			
			if (FD_ISSET(fwdent[i].listenfd, &rset)) {
				if (fwdent[i].proto == FWD_UDP) {
					do {
						bytes = recvfrom(fwdent[i].listenfd, dgrambuf, sizeof(dgrambuf),
						    0, &fwdent[i].sender, &fwdent[i].senderlen);
						if (bytes >= 0)
							bytesW = write(fwdent[i].dgramfd, dgrambuf, bytes);
					} while (bytes >= 0 && bytesW == bytes);
				} else
				if (fwdent[i].proto == FWD_TCP) {
				memset(&sa, 0, sizeof(sa));
                        	slen = sizeof(sa);
				if ( (newsock = accept(fwdent[i].listenfd, &sa, &slen)) != -1) {
					idx = find_free_slot(ud, cfg_max_users);
					if (idx == -1)
						close(newsock); /* handle this better */
					else {
                                        	ud[idx].u2s_i = ud[idx].u2s_o = ud[idx].u2s_buf;
                                        	ud[idx].s2u_i = ud[idx].s2u_o = ud[idx].s2u_buf;
                                        	ud[idx].connected = CONN_NO;
                                        	ud[idx].serv_fd = -1;
                                        	ud[idx].user_fd = newsock;
						ud[idx].fwdidx  = i;
						if (debug)
							fprintf(stderr, "connected to user on fd %d\n", newsock);
						ud[idx].serv_fd = setup_connect(SOCK_STREAM,
						    (struct sockaddr*)&fwdent[i].sin, sizeof(fwdent[i].sin), &result);
						if (debug)
							fprintf(stderr, "nonblocking connect returned %d\n", result);
						if (result == 1)
							ud[idx].connected = CONN_IN_PROG;
						else if (result == 0)
							ud[idx].connected = CONN_YES;
						else {
							/* syslog error */
							close_ud(&ud[idx]);
						}
					}
				}
				}
			}
				
		} 

		for (i = 0; i < cfg_max_users; i++) {
			if (ud[i].user_fd != -1) {

				if (ud[i].connected == CONN_YES) {

					/* Read Section */ 

					if (FD_ISSET(ud[i].serv_fd, &rset)) {
						bytes = read(ud[i].serv_fd, ud[i].s2u_i, &ud[i].s2u_buf[S2U_SIZE] - ud[i].s2u_i);
						if (bytes == -1) {
							if (errno != EWOULDBLOCK) {
								perror("server_read");
								close_ud(&ud[i]);
							}
							continue;
						} else if (bytes == 0) {
							close_ud(&ud[i]);
							continue;
						} else {
							ud[i].s2u_i += bytes;
							FD_SET(ud[i].user_fd, &wset); /* Try to write this data below */
						}
					}
                                        if (FD_ISSET(ud[i].user_fd, &rset)) {
                                                bytes = read(ud[i].user_fd, ud[i].u2s_i, &ud[i].u2s_buf[U2S_SIZE] - ud[i].u2s_i);
                                                if (bytes == -1) {
                                                        if (errno != EWOULDBLOCK) {
                                                                perror("server_read");
                                                                close_ud(&ud[i]);
                                                        }
                                                        continue;
                                                } else if (bytes == 0) {
                                                        close_ud(&ud[i]);
							continue;
                                                } else {
                                                        ud[i].u2s_i += bytes;
                                                        FD_SET(ud[i].serv_fd, &wset); /* Try to write this data below */
                                                }
        
                                	}

					/* Write section */
					
					if (FD_ISSET(ud[i].serv_fd, &wset) && ((bytes = ud[i].u2s_i - ud[i].u2s_o) > 0)) {
						bytesW = write(ud[i].serv_fd, ud[i].u2s_o, bytes);
						if (bytesW < 0) {
							if (errno != EWOULDBLOCK) {
								perror("server_write");
								close_ud(&ud[i]);
							}
							continue;
						} else {
							ud[i].u2s_o += bytesW;
							if (ud[i].u2s_o == ud[i].u2s_i)
								ud[i].u2s_o = ud[i].u2s_i = ud[i].u2s_buf;
						}
					}

                                        if (FD_ISSET(ud[i].user_fd, &wset) && ((bytes = ud[i].s2u_i - ud[i].s2u_o) > 0)) {
                                                bytesW = write(ud[i].user_fd, ud[i].s2u_o, bytes);
                                                if (bytesW < 0) {
                                                        if (errno != EWOULDBLOCK) {
                                                                perror("user_write");
                                                                close_ud(&ud[i]);
                                                        }
                                                        continue;
                                                } else {
                                                        ud[i].s2u_o += bytesW;
                                                        if (ud[i].s2u_o == ud[i].s2u_i)
                                                                ud[i].s2u_o = ud[i].s2u_i = ud[i].s2u_buf;
                                                }
                                        }


				} else if (ud[i].connected == CONN_IN_PROG) {
					if (FD_ISSET(ud[i].serv_fd, &rset) ||
					    FD_ISSET(ud[i].serv_fd, &wset)) {
						if (read(ud[i].serv_fd, fakebuf, 0) == -1) {
							close_ud(&ud[i]);
							continue;
						} else
							ud[i].connected = CONN_YES;
					}
				}
			}
		}
	}
}
